package nu.dll.md5integrity;

import java.util.*;
import java.util.regex.*;
import java.io.*;
import java.security.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

public class MD5Integrity {
    MessageDigest stdmd = null;
    Map filedata = new HashMap();
    File databaseFile = new File(System.getProperty("user.home"), ".md5database");
    ObjectOutputStream dbos = null;
    ObjectInputStream dbis = null;

    CheckActionCallback checkActionCallback = null;
    ProgressListener progressListener = null;

    boolean updatedb = false;
    boolean checkmtime = true;

    boolean syncDatabase = true;

    int failedCheckCount = 0;

    public MD5Integrity(boolean checkmtime) {
	this.checkmtime = checkmtime;
	initmd5();
    }

    public File getDatabase() {
	return databaseFile;
    }

    public void setDatabase(File f) {
	this.databaseFile = f;
    }

    public void setCheckActionCallback(CheckActionCallback cac) {
	this.checkActionCallback = cac;
    }

    public void setProgressListener(ProgressListener pl) {
	this.progressListener = pl;
    }

    void initOutputStream() throws IOException {
	dbos = new ObjectOutputStream(new GZIPOutputStream(new FileOutputStream(databaseFile)));
    }

    void initInputStream() throws IOException {
	dbis = new ObjectInputStream(new GZIPInputStream(new FileInputStream(databaseFile)));
    }

    public void initCheck() throws IOException {
	readFullDatabase();
	syncDatabase = false;
    }

    boolean aborted = false;
    public void check(File root) throws IOException {
	File[] children = root.listFiles();
	for (int i=0; !aborted && i < children.length; i++) {
	    if (children[i].isDirectory()) {
		check(children[i]);
	    } else {
		if (progressListener != null) {
		    progressListener.fileChecking(children[i]);
		}
		FileData oldFileData = getStoredFileData(children[i]);		
		boolean docheck = true;
		if (oldFileData == null && checkActionCallback != null) {
		    docheck = false;
		    byte[] curdigest = calcMD5(children[i]);
		    FileData newFileData = new FileData(children[i], curdigest);
		    int action = checkActionCallback.fileModified(null, newFileData);
		    if (action == CheckActionCallback.ACTION_UPDATEDB) {
			syncDatabase = true;
			filedata.put(newFileData.canonicalPath, newFileData);
		    }
		    if (action == CheckActionCallback.ACTION_ABORT) {
			aborted = true;
		    }
		    failedCheckCount++;
		}
		if (docheck) {
		    if (checkmtime && children[i].lastModified() == oldFileData.lastModified) {
			if (progressListener != null) {
			    progressListener.fileChecked(children[i], true);
			}
		    } else {
			byte[] curdigest = calcMD5(children[i]);
			if (Arrays.equals(curdigest, oldFileData.md5)) {
			    if (progressListener != null) {
				progressListener.fileChecked(children[i], true);
			    }
			} else {
			    if (progressListener != null) {
				progressListener.fileChecked(children[i], false);
			    }
			    failedCheckCount++;
			    if (checkActionCallback != null) {
				FileData newFileData = new FileData(children[i], curdigest);
				int action = checkActionCallback.fileModified(oldFileData, newFileData);
									      
				switch (action) {
				case CheckActionCallback.ACTION_UPDATEDB:
				    syncDatabase = true;
				    filedata.put(newFileData.canonicalPath, newFileData); 
				    break;
				case CheckActionCallback.ACTION_IGNORE:
				    filedata.remove(newFileData.canonicalPath);
				    break;
				case CheckActionCallback.ACTION_ABORT:
				    aborted = true;
				    break;
				default:
				    throw new IllegalArgumentException("Unknown callback retval: " + action);
				}
			    }
			}
		    }
		}
	    }
	}
    }

    public int getFailedCheckCount() {
	return failedCheckCount;
    }

    void closeis() throws IOException {
	dbis.close();
	dbis = null;
    }

    FileData getStoredFileData(File file) throws IOException {
	String cname = file.getCanonicalPath();
	return (FileData) filedata.get(cname);
    }

    void readFullDatabase() throws IOException {
	initInputStream();
	int filecount = 0;
	Object lastobject = null;
	try {
	    boolean reading = true;
	    while (reading) {
		lastobject = dbis.readObject();
		FileData fd = (FileData) lastobject;
		reading = fd != null;
		if (!reading) {
		    dprintln("EOF.");
		    continue;
		}
		filecount++;
		filedata.put(fd.canonicalPath, fd);
	    }
	} catch (EOFException ex1) {
	    // expected
	} catch (IOException ex1) {
	    dprintln("readFullDatabase(): " + ex1.toString());
	} catch (ClassNotFoundException ex2) {
	    throw new IOException("Invalid object in database: " + ex2.toString());
	} catch (ClassCastException ex3) {
	    throw new IOException("Bad class in database: " + lastobject.getClass().getName());
	}

	dprintln(filecount + " database entries read.");
	dbis.close();
	dbis = null;
    }

    public void initScan(boolean updatedb) throws IOException {
	this.updatedb = updatedb;
	if (updatedb) {
	    try {
		readFullDatabase();
	    } catch (FileNotFoundException ex1) {
		dprintln("initScan(" + updatedb + "): " + ex1.toString());
	    }
	}
    }

    void scan(File root) throws IOException {
	File[] children = root.listFiles();
	if (children == null) {
	    dprintln("null children in " + root);
	    return;
	}
	for (int i=0; !aborted && i < children.length; i++) {
	    if (children[i].isDirectory()) {
		scan(children[i]);
	    } else {
		if (progressListener != null) {
		    progressListener.fileChecking(children[i]);
		}
		String cname = children[i].getCanonicalPath();
		FileData oldData = null;
		if (updatedb && checkmtime) {
		    oldData = (FileData) filedata.get(cname);
		}
		if (oldData != null && oldData.lastModified == children[i].lastModified()) {
		    if (progressListener != null) {
			progressListener.fileScanned(children[i], oldData.md5);
		    }		    
		} else {
		    byte[] digest = calcMD5(children[i]);
		    saveData(children[i], digest);
		    if (progressListener != null) {
			progressListener.fileScanned(children[i], digest);
		    }
		}
	    }
	}
    }

    void sync() throws IOException {
	if (syncDatabase) {
	    if (dbos == null) initOutputStream();
	    if (filedata.size() > 0) {
		dprintln("Writing " + filedata.size() + " entries to database...");
	    }
	    for (Iterator i=filedata.values().iterator();i.hasNext();) {
		dbos.writeObject(i.next());
	    }
	}
	if (dbos != null) dbos.close();
	dbos = null;
    }

    void saveData(File file, byte[] digest) throws IOException {
	//dprintln(file + "\t" + hexstring(digest));
	FileData fd = new FileData(file, digest);
	if (dbos == null) {
	    initOutputStream();
	}
	dbos.writeObject(fd);
	filedata.remove(fd.canonicalPath);
    }

    byte[] md5buffer = new byte[512*1024];
    byte[] calcMD5(File file) throws IOException {
	stdmd.reset();
	InputStream is = new FileInputStream(file);
	int bytesRead;
	while((bytesRead = is.read(md5buffer)) != -1) {
	    stdmd.update(md5buffer, 0, bytesRead);
	}
	is.close();
	byte[] digest = stdmd.digest();
	return digest;
    }

    public static String hexstring(byte[] data) {
	StringBuffer buf = new StringBuffer();
	for (int i=0; i < data.length; i++) {
	    if (((int)  data[i] & 0xff) < 0x10) buf.append("0");
	    buf.append(Long.toString((int) data[i] & 0xff, 16));
	}
	return buf.toString();
    }

    void initmd5() {
	try {
	    stdmd = MessageDigest.getInstance("MD5");
	} catch (NoSuchAlgorithmException ex1) {
	    throw new RuntimeException(ex1.toString());
	}
    }

    boolean debug = Boolean.getBoolean("md5i.debug");
    void dprintln(String s) {
	if (debug) System.out.println(s);
    }

    static void usage() {
	System.out.println("");
	System.out.println("Usage:");
	System.out.println("");
	System.out.println("[-db <file>] [-v] [-nomtime] [[-updatedb] -scan <dir>] [[-alwaysignore] -check]");
	System.out.println("");
	System.out.println("Generic options:");
	System.out.println("\t -db FILE       uses FILE to store file data (default: ~/.md5database)");
	System.out.println("\t -v             verbose: outputs progress information");
	System.out.println("\t -nomtime       ignore file timestamps");
	System.out.println("");
	System.out.println("Scan options:");
	System.out.println("\t -scan DIR      scans DIR and all sub-directories");
	System.out.println("\t -updatedb      updates the database instead of creating a new");
	System.out.println("");
	System.out.println("Check uptions");
	System.out.println("\t -check DIR     checks all files in DIR and its sub-directories");
	System.out.println("\t -alwaysignore  Don't prompt on modified files: only report and ignore");
	System.out.println("");
	System.exit(255);
    }

    public static void main(String[] argv) throws Exception {
	List _args = Arrays.asList(argv);
	List args = new ArrayList();
	args.addAll(_args);
	if (args.size() < 2) {
	    System.out.println("Not enough arguments.");
	    usage();
	}

	boolean updatedb = false;
	boolean nomtime = false;
	final int ACTION_SCAN = 1;
	final int ACTION_CHECK = 2;
	int action = 0;
	String dir = null;
	final Set options = new HashSet();
	File databaseFile = null;
	for (Iterator i = args.iterator(); i.hasNext();) {
	    String arg = (String) i.next();
	    if (arg.startsWith("-")) {
		if (arg.equals("-scan")) {
		    action = ACTION_SCAN;
		    if (!i.hasNext()) usage();
		    dir = (String) i.next();
		} else if (arg.equals("-updatedb")) {
		    options.add(arg);
		} else if (arg.equals("-check")) {
		    action = ACTION_CHECK;
		    if (!i.hasNext()) usage();
		    dir = (String) i.next();
		} else if (arg.equals("-nomtime")) {
		    options.add(arg);
		} else if (arg.equals("-db")) {
		    if (!i.hasNext()) usage();
		    databaseFile = new File((String) i.next());
		} else if (arg.equals("-v") ||
			   arg.equals("-alwaysignore")) {
		    options.add(arg);
		} else {
		    System.out.println("Unkown argument: " + arg);
		    System.exit(0);
		}
	    }
	}

	File dirFile = new File(dir);
	if (!dirFile.exists() || !dirFile.isDirectory()) {
	    System.out.println("Directory not found: " + dirFile);
	    usage();
	}


	final BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));

	final boolean verbose = options.contains("-v");
	MD5Integrity md5i = new MD5Integrity(!options.contains("-nomtime"));

	if (databaseFile != null) {
	    md5i.setDatabase(databaseFile);
	}

	if (verbose) {
	    System.out.println("Using database: " + md5i.getDatabase().getCanonicalPath());
	}


	md5i.setProgressListener(new ProgressListener() {
		public void fileChecking(File f) {}
		public void fileScanned(File f, byte[] data) {
		    if (verbose) {
			System.out.println(f.toString() + ": " + hexstring(data));
		    }
		}
		public void fileChecked(File f, boolean ok) {
		    if (ok) {
			if (!verbose) return;
			System.out.print("[OK    ] ");
		    } else {
			System.out.print("[NOT OK] ");			
		    }
		    System.out.println(f);
		}
	    });
	switch (action) {
	case ACTION_SCAN:
	    md5i.initScan(options.contains("-updatedb"));
	    md5i.scan(dirFile);
	    md5i.sync();
	    break;
	case ACTION_CHECK:
	    md5i.setCheckActionCallback(new CheckActionCallback() {
		    public int fileModified(FileData oldData, FileData newData) {
			if (options.contains("-alwaysignore")) { 
			    return CheckActionCallback.ACTION_IGNORE;
			}
			if (oldData != null) {
			    System.out.println("Old MD5: " + hexstring(oldData.md5));
			    System.out.println("New MD5: " + hexstring(newData.md5));
			    System.out.println("Old timestamp: " + new Date(oldData.lastModified));
			    System.out.println("New timestamp: " + new Date(newData.lastModified));
			} else {
			    System.out.println("New file: " + newData.canonicalPath);
			    System.out.println("MD5: " + hexstring(newData.md5));
			    System.out.println("Timestamp: " + new Date(newData.lastModified));
			}
			while (true) {
			    System.out.print("Action: I=ignore, U=update database (default: I)?  ");
			    String action = null;
			    try {
				action = stdin.readLine().toUpperCase();
			    } catch (IOException ex1) {
				throw new RuntimeException(ex1.toString());
			    }
			    if (action.equals("U")) {
				return CheckActionCallback.ACTION_UPDATEDB;
			    } else if (action.equals("") || action.equals("I")) {
				return CheckActionCallback.ACTION_IGNORE;
			    }
			}
		    }
		});
	    try {
		md5i.initCheck();
	    } catch (FileNotFoundException ex1) {
		System.out.println("Database not found: " + ex1.getMessage());
		System.out.println("Use -scan to create one.");
		System.exit(255);
	    }
	    md5i.check(dirFile);
	    md5i.sync();
	    if (md5i.getFailedCheckCount() > 0) {
		if (verbose) {
		    System.out.println(md5i.getFailedCheckCount() + " files failed MD5 check.");
		}
		System.exit(1);
	    }
	    break;
	default:
	    System.out.println("Error: no action.");
	    break;
	}
    }
}
